use std::{
    hash::{DefaultHasher, Hash, Hasher},
    ops::Deref,
};

use bytes::Bytes;
use leptos::*;
use uuid::Uuid;

use crate::{
    api::{stream_chat, ChannelMessage, ChatMessage, ChatRole},
    app::components::{svgs::*, Card},
};

#[island]
fn ChatMessageBubble(
    msg: RwSignal<ChatMessage>,
    history: RwSignal<Vec<RwSignal<ChatMessage>>>,
) -> impl IntoView {
    let (edit_mode, set_edit_mode) = create_signal(false);
    let _ = create_effect(move |_| tracing::debug!(edit_mode=%edit_mode()));
    let role = match msg.get().role {
        ChatRole::System => "System",
        ChatRole::User => "User",
        ChatRole::Assistant => "Assistant",
    };

    let chat_class = match msg.get().role {
        ChatRole::System => "chat-start",
        ChatRole::Assistant => "chat-start",
        ChatRole::User => "chat-end",
    };

    let textarea_ref = NodeRef::<html::P>::new();

    use pulldown_cmark;
    let editable_p = move || {
        // TODO Convert back to raw str when editable
        let mode = if edit_mode.get() { "true" } else { "false" };
        let msg_str = move || msg.get().content.clone();
        let md_str = move || {
            let owned_str = msg_str();
            let parser = pulldown_cmark::Parser::new(&owned_str);
            let mut md_output = String::new();
            pulldown_cmark::html::push_html(&mut md_output, parser);

            md_output
        };

        let inner_p = move || {
            if edit_mode.get() {
                view! {

                    <p inner_html=move || { msg_str() }></p>
                }
            } else {
                view! {
                    <p inner_html=move || { md_str() }></p>
                }
            }
        };

        view! {
            <p node_ref=textarea_ref contenteditable=move || { mode }>

                {inner_p}
            </p>
        }
    };

    let enter_edit_mode = move |_ev| {
        set_edit_mode.set(true);

        textarea_ref().expect("").deref().clone().focus().unwrap();
    };

    let exit_edit_mode = move |_ev| {
        let inner_text = textarea_ref().expect("").deref().clone().inner_text();

        msg.update(move |msg| msg.content = inner_text);

        set_edit_mode.set(false);
    };

    let trash_message = move || {
        let id = msg.get().id;
        history.update(move |his| {
            his.retain(|elem| elem.get().id != id);
        });
    };

    let edit_button = move || {
        if edit_mode() {
            view! {
                <button
                    on:click=exit_edit_mode
                    class="absolute top-2 right-2 btn btn-xs btn-ghost btn-circle"
                >
                    <CheckmarkSvg/>
                </button>
            }
            .into_view()
        } else {
            view! {
                <button
                    on:click=enter_edit_mode
                    class="absolute top-2 right-2 btn btn-xs btn-ghost btn-circle"
                >
                    <PencilSvg/>
                </button>
            }
            .into_view()
        }
    };

    let trash_button = move || {
        if msg().role != ChatRole::System && !edit_mode() {
            view! {
                <button
                    on:click=move |_e| trash_message()
                    class="absolute right-2 bottom-2 btn btn-xs btn-ghost btn-circle"
                >
                    <TrashSvg/>
                </button>
            }
            .into_view()
        } else {
            "".into_view()
        }
    };

    view! {
        <div class=format!("chat {chat_class}")>
            <div class="chat-header">
                <div class="flex flex-nowrap">
                    <p class="grow">{role}</p>
                </div>
            </div>
            <div class="chat-bubble">
                {edit_button} {trash_button} <div class="flex flex-nowrap w-full h-full">
                    <div>{editable_p}</div>
                    // as the edit/trash buttons are absolutely positions this makes sure they dont overlap with text
                    // TODO move them here ?
                    <div class="w-6 h-12 shrink-0">""</div>
                </div>
            </div>
        </div>
    }
}

#[component]
pub fn ChatHistory(
    history: RwSignal<Vec<RwSignal<ChatMessage>>>,
    response: RwSignal<String>,
    receiving_response: RwSignal<bool>,
) -> impl IntoView {
    let streaming_response_elem = move || {
        if receiving_response() {
            view! {
                <div class="chat chat-start">
                    <div class="chat-header">"Assistant"</div>
                    <div class="chat-bubble">
                        <span inner_html=response></span>
                        <span class="loading loading-ball loading-xs"></span>
                    </div>
                </div>
            }
            .into_view()
        } else {
            view! { <></> }.into_view()
        }
    };

    view! {
        <div class="flex flex-col flex-nowrap p-2">
            <For
                each=move || history().into_iter()
                key=|entry| {
                    let mut s = DefaultHasher::new();
                    entry.get().hash(&mut s);
                    s.finish()
                }

                children=move |msg| {
                    view! { <ChatMessageBubble msg history/> }
                }
            />

            {streaming_response_elem}
        </div>
    }
}

#[component]
fn ChatPrompt(
    receiving_response: RwSignal<bool>,
    history: RwSignal<Vec<RwSignal<ChatMessage>>>,
    response: RwSignal<String>,
) -> impl IntoView {
    let stream_completion = create_action(move |id: &Uuid| {
        use futures_util::StreamExt;

        let id = *id;
        let history_inner = history();

        let handle_stop = move || {
            history.update(move |vec| {
                let msg = ChatMessage::new(ChatRole::Assistant, response.get_untracked());
                vec.push(create_rw_signal(msg));
            });
            response.set("".to_string());
            receiving_response.set(false);
        };
        let handle_token = move |string: String| {
            response.update(move |resp| {
                resp.push_str(string.as_str());
            });
        };

        let handle_chunk = move |chunk: Bytes| {
            tracing::debug!(?chunk);
            let deserializer = serde_json::Deserializer::from_slice(&chunk);
            for msg in deserializer.into_iter::<ChannelMessage>() {
                tracing::debug!(?msg);
                match msg {
                    Ok(ChannelMessage::Stop) => handle_stop(),
                    Ok(ChannelMessage::Token(string)) => handle_token(string),
                    Err(err) => {
                        tracing::error!(?err);
                        panic!("Unknown stream errror")
                    }
                };
            }
        };

        async move {
            let his_inner = history_inner
                .into_iter()
                .map(|sig| sig.get())
                .collect::<Vec<_>>();
            let res_stream = stream_chat(id, his_inner).await;
            if let Ok(stream) = res_stream {
                let mut stream = stream.into_inner();

                tracing::debug!("Stream start");
                while let Some(Ok(chunk)) = stream.next().await {
                    handle_chunk(chunk);
                }
                tracing::debug!("Stream end");
            }
        }
    });

    let prompt_input_element: NodeRef<html::Input> = create_node_ref();

    let submit_chat = move || {
        receiving_response.set(true);

        let input = prompt_input_element().expect("<input> should be mounted");
        let value = input.value();
        input.set_value("");
        if !value.is_empty() {
            let new_chat_entry = ChatMessage::new(ChatRole::User, value.clone());
            let new_entry_sig = create_rw_signal(new_chat_entry);
            history.update(move |hist| hist.push(new_entry_sig));
        }
        stream_completion.dispatch(Uuid::now_v7());
    };
    let submit_on_enter = move |ev: web_sys::KeyboardEvent| {
        const ENTER_KEY: u32 = 13;
        let key_code = ev.key_code();
        let shift_key = ev.shift_key();
        if key_code == ENTER_KEY && !shift_key {
            submit_chat()
        }
    };

    let refresh_last_message = move |ev: web_sys::MouseEvent| {
        ev.prevent_default();
        receiving_response.set(true);

        history.update(move |hist| {
            hist.pop();
        });

        stream_completion.dispatch(Uuid::now_v7());
    };

    view! {
        <div class="flex-nowrap justify-around p-2 card-actions">
            <input
                class="w-full input grow input-bordered input-primary"
                type="text"
                placeholder="Type here"
                node_ref=prompt_input_element
                on:keyup=submit_on_enter
            />

            <button
                type="submit"
                class="btn btn-primary"
                class:btn-disabled=receiving_response
                on:click=move |ev| {
                    ev.prevent_default();
                    submit_chat()
                }
            >

                <SendSvg/>
            </button>
            <button
                on:click=refresh_last_message
                class="btn btn-ghost"
                class:btn-disabled=move || {
                    receiving_response()
                        || history().last().unwrap().get().role != ChatRole::Assistant
                }
            >

                <RefreshSvg/>
            </button>
        </div>
    }
}

#[component]
pub fn ChatPage() -> impl IntoView {
    let initial_history = vec![create_rw_signal(ChatMessage::new(
        ChatRole::System,
        "You are an helpful AI-based assistant.".to_string(),
    ))];

    let history = create_rw_signal(initial_history);
    let response = create_rw_signal("".to_string());
    let receiving_response = create_rw_signal(false);

    view! {
        <Card>
            <div class="flex flex-col justify-between h-full">
                <ChatHistory receiving_response history response/>
                <ChatPrompt receiving_response history response/>
            </div>
        </Card>
    }
}
