use async_trait::async_trait;
use serde::{Deserialize, Serialize};
use tokio::sync::mpsc;

use crate::api::{ChannelMessage, ChatMessage};
pub mod llama_chat;
pub mod llama_completion;
pub mod runner;

// struct SamplerSettings {
//     temp:  i64,
//     top_k: i64,
//     top_p: i64,
//     // ...
// }

#[async_trait]
pub trait ChatService: Sized {
    async fn chat_request(self, history: Vec<ChatMessage>, rx: mpsc::Sender<ChannelMessage>);
}

#[derive(Debug, Serialize, Deserialize, Clone, PartialEq)]
pub enum BackendServiceStatus {
    Started,
    Stopped,
}

use std::sync::Arc;

use sysinfo::{Signal, System};
use tokio::{sync::Mutex, task::JoinHandle};

use super::backends::runner::{Runner, RunnerArgs};

fn kill_child_procs(child_str: &'static str) {
    let mut system = System::new();
    system.refresh_processes();

    let current_pid = sysinfo::get_current_pid().unwrap();
    let current_process = system.process(current_pid).unwrap();

    terminate_child_processes(child_str, &system, current_process);
}

fn collect_ape_child_processes(
    child_str: &'static str,
    system: &System,
    pid: sysinfo::Pid,
) -> Vec<sysinfo::Pid> {
    let mut child_pids = Vec::new();

    for process in system.processes().values() {
        if let Some(parent_pid) = process.parent() {
            if process.name().contains(child_str) && is_ancestor(system, parent_pid, pid) {
                tracing::debug!(msg = "Found process to kill", ?process);
                child_pids.push(process.pid());
            }
        }
    }
    child_pids.sort();
    child_pids.reverse();

    child_pids
}

fn is_ancestor(system: &System, pid: sysinfo::Pid, ancestor_pid: sysinfo::Pid) -> bool {
    if pid == ancestor_pid {
        return true;
    }

    if let Some(process) = system.process(pid) {
        if let Some(parent_pid) = process.parent() {
            return is_ancestor(system, parent_pid, ancestor_pid);
        }
    }

    false
}

fn terminate_child_processes(child_str: &'static str, system: &System, process: &sysinfo::Process) {
    let child_pids = collect_ape_child_processes(child_str, system, process.pid());

    for child_pid in child_pids {
        if let Some(child_process) = system.process(child_pid) {
            // terminate_child_processes(system, child_process);
            if let Some(res) = child_process.kill_with(Signal::Term) {
                tracing::debug!(
                    "Sent SIGTERM to child process with PID: {} {}",
                    child_pid,
                    res
                );
            } else {
                tracing::debug!(
                    "Failed to send SIGTERM to child process with PID: {}",
                    child_pid
                );
            }
        }
    }
}

#[derive(Debug, Clone)]
pub struct BackendService {
    status: Arc<Mutex<BackendServiceStatus>>,
    task_handle: Arc<Mutex<Option<JoinHandle<anyhow::Result<()>>>>>,
}

impl Default for BackendService {
    fn default() -> Self {
        Self::new()
    }
}

impl BackendService {
    pub fn new() -> Self {
        Self {
            status: Arc::new(Mutex::new(BackendServiceStatus::Stopped)),
            task_handle: Arc::new(Mutex::new(None)),
        }
    }

    #[tracing::instrument(level = "debug")]
    pub async fn start(&self) -> anyhow::Result<()> {
        let mut status = self.status.clone().lock_owned().await;
        let mut task_handle = self.task_handle.clone().lock_owned().await;

        if task_handle.is_none() && *status == BackendServiceStatus::Stopped {
            let args = RunnerArgs::new(
                4096,
                9999,
                // "./models/phi-3-mini-4k-instruct-imat-Q8_0.gguf".to_string(),
                "./models/meta-llama-3-8b-instruct-imat-IQ3_XXS.gguf".to_string(),
            );
            let starter_fut = Runner::new_llama_server_bin(args).run();
            *task_handle = Some(tokio::task::spawn(starter_fut));
            *status = BackendServiceStatus::Started;
        }

        Ok(())
    }

    #[tracing::instrument(level = "debug")]
    pub async fn stop(&self) -> anyhow::Result<()> {
        let mut status = self.status.clone().lock_owned().await;
        let mut task_handle = self.task_handle.clone().lock_owned().await;

        if task_handle.is_some() && *status == BackendServiceStatus::Started {
            tracing::warn!(msg = "Aborting Process", ?task_handle);
            task_handle.as_mut().unwrap().abort();

            // FIXME
            // this is an ugly workaround as we are starting llamafile in a bash shell which somehow does not propagate systerm correctly
            // we should probably keep track of our child process pids and selectively kill stuff
            tracing::warn!(msg = "Killing Childs", ?task_handle);
            kill_child_procs(".ape-");
            kill_child_procs("llama-server");

            *task_handle = None;
            *status = BackendServiceStatus::Stopped;
        }

        Ok(())
    }

    pub async fn status(&self) -> BackendServiceStatus {
        self.status.lock().await.clone()
    }
}

#[cfg(test)]
mod tests {
    use std::time::Duration;

    use tokio::{sync::mpsc, time::sleep};
    use uuid::Uuid;

    use crate::{
        api::{ChannelMessage, ChatMessage, ChatRole},
        server::backends::{
            llama_chat::LlamaService, BackendService, BackendServiceStatus, ChatService,
        },
    };

    #[tracing_test::traced_test]
    #[tokio::test]
    async fn it_can_be_started_then_queried_and_then_stopped() {
        let service_handle = BackendService::new();

        assert_eq!(service_handle.status().await, BackendServiceStatus::Stopped);

        service_handle.start().await;
        assert_eq!(service_handle.status().await, BackendServiceStatus::Started);

        let history = vec![
            ChatMessage::new(
                ChatRole::System,
                "You are an helpful AI-based assistant.".to_string(),
            ),
            ChatMessage::new(ChatRole::User, "What is 2+2 ?".to_string()),
        ];

        let (tx, mut rx) = mpsc::channel::<ChannelMessage>(1);
        let llama_service = LlamaService::new(Uuid::now_v7());
        sleep(Duration::from_secs(30)).await;
        let chat_req = llama_service.chat_request(history, tx);
        tokio::task::spawn(chat_req);
        sleep(Duration::from_millis(1000)).await;

        let mut response = String::new();
        while let Some(msg) = rx.recv().await {
            if let ChannelMessage::Token(token) = msg {
                response.push_str(token.as_str());
            }
        }

        tracing::debug!("response: {}", response);
        assert!(response.contains('4'));

        service_handle.stop().await;

        assert_eq!(service_handle.status().await, BackendServiceStatus::Stopped);
    }
}
