use std::{
    future::Future,
    pin::Pin,
    sync::Arc,
    task::{Context, Poll},
};

use std::sync::Once;

use axum::{body::Body, http::Request};
use pin_project_lite::pin_project;
use tower::{Layer, Service};
use uuid::Uuid; // Make sure to include `uuid` crate in your Cargo.toml

pub struct LoggingLayer;

impl<S> Layer<S> for LoggingLayer {
    type Service = LoggingService<S>;

    fn layer(&self, inner: S) -> Self::Service {
        LoggingService { inner }
    }
}

pub struct LoggingService<T> {
    inner: T,
}

impl<T> Service<Request<Body>> for LoggingService<T>
where
    T: Service<Request<Body>>,
{
    type Error = T::Error;
    type Future = LoggingServiceFuture<T::Future>;
    type Response = T::Response;

    fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
        self.inner.poll_ready(cx)
    }

    fn call(&mut self, req: Request<Body>) -> Self::Future {
        let request_uuid = Uuid::now_v7(); // Generate UUID v7

        tracing::info!("[{}] {} {}", request_uuid, req.method(), req.uri());
        for (k, v) in req.headers() {
            tracing::info!(
                "[{}] {}: {}",
                request_uuid,
                k,
                v.to_str().unwrap_or("invalid header")
            );
        }

        LoggingServiceFuture {
            inner: self.inner.call(req),
            uuid: Arc::new(request_uuid), // Store UUID in an Arc for shared ownership
        }
    }
}

pin_project! {
    pub struct LoggingServiceFuture<T> {
        #[pin]
        inner: T,
        uuid: Arc<Uuid>, // Shared state between LoggingService and LoggingServiceFuture
    }
}

impl<T> Future for LoggingServiceFuture<T>
where
    T: Future,
{
    type Output = T::Output;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let this = self.project();
        match this.inner.poll(cx) {
            Poll::Pending => Poll::Pending,
            Poll::Ready(output) => {
                tracing::trace!("[{}] request finished", this.uuid);
                Poll::Ready(output)
            }
        }
    }
}

static INIT: Once = Once::new();

pub fn initialize_logger() {
    INIT.call_once(|| {
        let env_filter = tracing_subscriber::EnvFilter::builder()
            .with_default_directive(tracing::level_filters::LevelFilter::INFO.into())
            .from_env_lossy();

        tracing_subscriber::fmt()
            .compact()
            .with_env_filter(env_filter)
            .init();
    });
}
