use crate::{
    config::ModelSpec, error::AppError, inference_process::InferenceProcess, state::AppState,
    util::parse_size,
};
use axum::{
    body::Body,
    http::{Request, Response},
    response::IntoResponse,
};
use reqwest::Client;
use reqwest_middleware::{ClientBuilder, ClientWithMiddleware};
use reqwest_retry::{policies::ExponentialBackoff, RetryTransientMiddleware};

pub async fn proxy_request(
    req: Request<Body>,
    spec: &ModelSpec,
) -> Result<Response<Body>, AppError> {
    let retry_policy = ExponentialBackoff::builder()
        .retry_bounds(
            std::time::Duration::from_millis(500),
            std::time::Duration::from_secs(8),
        )
        .jitter(reqwest_retry::Jitter::None)
        .base(2)
        .build_with_max_retries(8);

    let client: ClientWithMiddleware = ClientBuilder::new(Client::new())
        .with(RetryTransientMiddleware::new_with_policy(retry_policy))
        .build();

    let internal_port = spec.internal_port.expect("Internal port must be set");
    let uri = format!(
        "http://127.0.0.1:{}{}",
        internal_port,
        req.uri()
            .path_and_query()
            .map(|pq| pq.as_str())
            .unwrap_or("")
    );

    let mut request_builder = client.request(req.method().clone(), &uri);
    for (name, value) in req.headers() {
        request_builder = request_builder.header(name, value);
    }

    let max_size = parse_size("1G").unwrap() as usize;
    let body_bytes = axum::body::to_bytes(req.into_body(), max_size).await?;
    let request = request_builder.body(body_bytes).build()?;
    let response = client.execute(request).await?;

    let mut builder = Response::builder().status(response.status());
    for (name, value) in response.headers() {
        builder = builder.header(name, value);
    }

    let byte_stream = response.bytes_stream();
    let body = Body::from_stream(byte_stream);
    Ok(builder.body(body)?)
}

pub async fn handle_request(
    req: Request<Body>,
    spec: ModelSpec,
    state: AppState,
) -> impl IntoResponse {
    let _ = InferenceProcess::get_or_spawn(&spec, state).await?;
    let response = proxy_request(req, &spec).await?;
    Ok::<Response<Body>, AppError>(response)
}
